

#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import math
import gtk

_LINE_WIDTH = 2

class Pkg:
    ARC_RADIUS   = 10
    TEXT_PADDING = 4

    def __init__ (self, name, requires = [], required_by = [], ctx = None):
        self.name = name
        
        self.requires    = requires
        self.required_by = required_by

        self.set_position (None, None)
        self.set_context (ctx)

    def copy (self):
        return Pkg (self.name, self.requires, self.required_by)

    def _calc_text_extents (self):
        """Calculate the bounding rectangle of the glyphs on-screen
        if the package's name is drawn at the origin.
        """
        (x, y, width, height, dx, dy) = self._ctx.text_extents (self.name)

        self._text_x      = x
        self._text_y      = y
        self._text_width  = width
        self._text_height = height

    def _calc_size (self):
        """Calculate the bounding dimensions of the graphic and text
        representing the package.
        """
        self.width  = max (self._text_width, 2.0 * (_LINE_WIDTH + self.ARC_RADIUS))
        self.height = self._text_height + self.TEXT_PADDING + 2.0 * (_LINE_WIDTH + self.ARC_RADIUS)

    def _calc_left_cnx (self):
        """Calculate the position of the left connection point, relative
        to the package's origin.
        """
        self.lcx = (self.width / 2.0) - self.ARC_RADIUS
        self.lcy = self.height - (self.ARC_RADIUS + _LINE_WIDTH)
    
    def _calc_right_cnx (self):
        """Calculate the position of the right connection point, relative
        to the package's origin.
        """
        self.rcx = (self.width / 2.0) + self.ARC_RADIUS
        self.rcy = self.height - (self.ARC_RADIUS + _LINE_WIDTH)

    def reset_context (self):
        """Reset all context dependent state."""
        self._ctx = None

        self.x      = 0.0
        self.y      = 0.0
        self.width  = 0.0
        self.height = 0.0

        self._text_x      = 0.0
        self._text_y      = 0.0
        self._text_width  = 0.0
        self._text_height = 0.0

        self.lcx  = 0.0
        self.lcy  = 0.0
        self.rcx = 0.0
        self.rcy = 0.0

    def set_context (self, ctx):
        """Set the cairo context with which the package will be drawn
        and calculate all context dependent state.
        """
        self.reset_context ()

        if ctx is None:
            return

        self._ctx = ctx

        self._calc_text_extents ()
        self._calc_size ()
        self._calc_left_cnx ()
        self._calc_right_cnx ()

    def set_position (self, x, y):
        """Configure the package to be drawn at position (@x, @y)."""
        self.x = x
        self.y = y

    def draw (self):
        """Draw the graphic and text representing a package."""
        assert not self.x is None
        assert not self.y is None
        assert not self._ctx is None
        
        self._ctx.save ()
        self._ctx.translate (self.x + (self.width / 2.0),
                             self.y + self.height - (self.ARC_RADIUS + _LINE_WIDTH))

        self._ctx.new_path ()
        self._ctx.arc (0, 0, self.ARC_RADIUS, 0, 2.0 * math.pi)
        self._ctx.stroke ()
        
        self._ctx.restore ()
        
        if self._text_width < self.width:
            center_x = (self.width - self._text_width) / 2.0
        else:
            center_x = 0.0

        self._ctx.move_to (self.x + center_x - self._text_x, self.y - self._text_y)
        self._ctx.show_text (self.name)

class Connector:
    def __init__ (self, pkg1, pkg2, ctx = None):
        self._pkg1 = pkg1
        self._pkg2 = pkg2
        
        self.set_context (ctx)

    def reset_context (self):
        self._ctx = None

    def set_context (self, ctx):
        self.reset_context ()

        if ctx is None:
            return

        self._ctx = ctx

    def draw (self):
        assert not self._pkg1.x is None
        assert not self._pkg1.y is None
        assert not self._pkg2.x is None
        assert not self._pkg2.y is None
        assert not self._ctx is None

        if self._pkg2.x > self._pkg1.x:
            cnx1 = (self._pkg1.rcx, self._pkg1.rcy)
            cnx2 = (self._pkg2.lcx, self._pkg2.lcy)
        else:
            cnx1 = (self._pkg1.lcx, self._pkg1.lcy)
            cnx2 = (self._pkg2.rcx, self._pkg2.rcy)
        
        self._ctx.new_path ()
        self._ctx.move_to (self._pkg1.x + cnx1[0],
                           self._pkg1.y + cnx1[1])
        self._ctx.line_to (self._pkg2.x + cnx2[0],
                           self._pkg2.y + cnx2[1])
        self._ctx.stroke ()

class DepsWindow:
    def __init__ (self, packages, selected):
        self._packages = packages
        self._selected = selected
        self._back     = []
        self._forward  = []
        
        self._window = gtk.Window ()
        self._window.set_title (self._selected)
        self._window.connect ("delete-event", gtk.main_quit)

        self._vbox = gtk.VBox ()
        self._window.add (self._vbox)
        self._vbox.show ()

        self._setup_toolbar ()
        self._setup_drawingarea ()

        self._update_ui ()

    def _setup_toolbar (self):
        ui_string = '''
<ui>
 <toolbar>
  <toolitem action="Back" />
  <toolitem action="Forward" />
 </toolbar>
</ui>
'''
        actions = [
            ( "Back",    gtk.STOCK_GO_BACK,    _("_Back"),    "<Alt>leftarrow",  _("Go back to previous package"), self._handle_back),
            ( "Forward", gtk.STOCK_GO_FORWARD, _("_Forward"), "<Alt>rightarrow", _("Go forward to next package"),  self._handle_forward),
        ]

        action_group = gtk.ActionGroup ("WindowActions")
        action_group.add_actions (actions)

        self._ui_manager = gtk.UIManager ()
        self._ui_manager.insert_action_group (action_group, 0)
        self._ui_manager.connect ("add-widget", self._add_widget)
        self._ui_manager.add_ui_from_string (ui_string)
        self._ui_manager.ensure_update ()

        self._window.add_accel_group (self._ui_manager.get_accel_group ())

        self._back_action    = action_group.get_action ("Back")
        self._forward_action = action_group.get_action ("Forward")

    def _add_widget (self, ui_maanger, widget):
        self._vbox.pack_start (widget, False, False, 0)

    def _update_ui (self):
        self._back_action.set_sensitive    (len (self._back) > 0)
        self._forward_action.set_sensitive (len (self._forward) > 0)
        
        self._window.set_title (self._selected)

        self._left_pkgs  = []
        self._right_pkgs = []
        
        self._middle_pkg = self._packages[self._selected].copy ()

        for pkg_name in self._middle_pkg.required_by:
            self._left_pkgs.append (self._packages[pkg_name].copy ())
        
        for pkg_name in self._middle_pkg.requires:
            self._right_pkgs.append (self._packages[pkg_name].copy ())
        
        self._drawingarea.queue_draw ()
        
    def _select (self, name):
        if name == self._selected:
            return
        
        self._forward = []
        self._back.append (self._selected)
        self._selected = name
        
        self._update_ui ()

    def _handle_back (self, action):
        name = self._back.pop ()
        self._forward.append (self._selected)
        self._selected = name
        self._update_ui ()

    def _handle_forward (self, action):
        name = self._forward.pop ()
        self._back.append (self._selected)
        self._selected = name
        self._update_ui ()

    def _setup_drawingarea (self):
        self._drawingarea = gtk.DrawingArea ()
        self._drawingarea.connect ("expose-event", self._handle_expose)
        self._drawingarea.set_size_request (640, 480)
        self._drawingarea.show ()

        self._drawingarea.add_events (gtk.gdk.BUTTON_PRESS_MASK)
        self._drawingarea.connect ("button-press-event", self._handle_button_press)
        
        self._vbox.pack_start (self._drawingarea, True, False, 0)

    def _draw_pkgs (self, ctx, pkgs, connect_to, center, vpad):
        total_height = 0
        for pkg in pkgs:
            pkg.set_context (ctx)
            total_height += pkg.height
        if len (pkgs) > 1:
            total_height += (len (pkgs) - 1) * vpad
    
        y = center[1] - (total_height / 2.0)

        for pkg in pkgs:
            pkg.set_position (center[0] - (pkg.width / 2.0), y)
            pkg.draw ()

            c = Connector (pkg, connect_to, ctx).draw ()

            y += pkg.height + vpad

    def _draw (self, ctx, center, pkg, left_pkgs, right_pkgs):
        pkg.set_context (ctx)
        pkg.set_position (center[0] - (pkg.width / 2.0),
                          center[1] - (pkg.height / 2.0))
        pkg.draw ()
        
        self._draw_pkgs (ctx, left_pkgs,  pkg, (center[0] / 2.0,               center[1]), 10)
        self._draw_pkgs (ctx, right_pkgs, pkg, (center[0] + (center[0] / 2.0), center[1]), 10)
    
    def _handle_expose (self, widget, event):
        ctx = widget.window.cairo_create ()

        ctx.rectangle (0, 0,
                       widget.allocation.width,
                       widget.allocation.height)
        ctx.set_source_rgb (1.0, 1.0, 1.0)
        ctx.fill ()
        
        ctx.set_source_rgb (0.0, 0.0, 0.0)
        ctx.set_line_width (_LINE_WIDTH)
        ctx.select_font_face ('sans-serif')
        ctx.set_font_size (20)

        self._draw (ctx,
                    (320, 240),
                    self._middle_pkg,
                    self._left_pkgs,
                    self._right_pkgs)

    def _handle_button_press (self, widget, event):
        def find_pkg (pkgs, x, y):
            for pkg in pkgs:
                if pkg.x is None or pkg.y is None:
                    continue

                if x >= pkg.x and x < (pkg.x + pkg.width) and \
                   y >= pkg.y and y < (pkg.y + pkg.height):
                    return pkg
            return None

        pkg = find_pkg (self._left_pkgs, event.x, event.y)
        if pkg is None:
            pkg = find_pkg (self._right_pkgs, event.x, event.y)

        if not pkg is None:
            self._select (pkg.name)
    
    def show (self):
        self._window.show ()
